import abc
import typing as t

from pybus.core.api.typing import HandlerType, MessageType, ReturnType
from pybus.core.types import EMPTY


@t.runtime_checkable
class HandlerProtocol(t.Protocol):
    """Handler protocol for runtime checking."""

    async def handle(self, message: MessageType) -> ReturnType:
        """Handle message."""

    async def add_event(self, event: MessageType) -> None:
        """Add event to emit later."""

    async def dump_events(self) -> list[MessageType]:
        """Return list of collected events."""


class AbstractHandler(t.Generic[ReturnType], metaclass=abc.ABCMeta):
    """Handler protocol."""

    @abc.abstractmethod
    async def handle(self, message: MessageType) -> ReturnType:
        """Handle message."""

    @abc.abstractmethod
    async def add_event(self, event: MessageType) -> None:
        """Add event to emit later."""

    @abc.abstractmethod
    async def dump_events(self) -> list[MessageType]:
        """Return list of collected events."""


class AbstractHandlerWrapper(
    AbstractHandler[ReturnType],
    t.Generic[ReturnType],
    metaclass=abc.ABCMeta,
):
    """Handler wrapper."""

    def __init__(
        self,
        meta: "HandlerMetaDataProtocol",
        **initkwargs,
    ):
        self._handler = meta.handler
        self._message = meta.message
        self._inject = meta.inject
        self._argname = meta.argname

        self._initkwargs = initkwargs
        self._events = []


class HandlerMetaDataProtocol(t.Protocol):
    """Meta data for dispatcher"""
    handler: HandlerType
    inject: bool
    argname: t.Optional[str] = EMPTY
    message: t.Optional[MessageType] = EMPTY



from typing import Protocol, runtime_checkable

import pytest

from pybus.core import exceptions as exc
from pybus.core.inspection import sig, utils


class BadAsync:
    def handle(self):
        pass


class BadNoMethod:
    pass


class BadSignature:
    async def handle(self):
        pass


class Good:
    async def handle(self, message: str):
        pass


async def handle(message: str):
    pass


class BadAssigned:
    pass


BadAssigned.handle = handle


@pytest.fixture(scope="module")
def protocol():
    @runtime_checkable
    class _Protocol(Protocol):
        async def handle(self, message: str):
            pass

    return _Protocol


async def handle(message: str):
    pass




@pytest.mark.parametrize(
    "message, handler, argname, kind, exception",
    (
        ("on_startup", None, "message", "pos", exc.HandlerDoesNotExist),
        ("on_startup", handle, None, "pos_or_kw", exc.ImproperlyConfigured),
        ("on_startup", handle, None, "kw_only", exc.ImproperlyConfigured),
        ("on_startup", handle, None, "kw", exc.ImproperlyConfigured),
        ("on_startup", handle, "message", "badkind", exc.ImproperlyConfigured),
    )
)
def test_check_signature_fails(message, handler, argname, kind, exception):
    with pytest.raises(exception):
        sig.check_signature(message, handler, argname, kind=kind)


@pytest.mark.parametrize(
    "cls, expected",
    (
        (BadAsync, False),
        (BadNoMethod, False),
        (Good, True),
        (BadSignature, False),
        (handle, False),
        (BadAssigned, False),
    )
)
def test_implements_protocol(cls, expected, protocol):
    result = utils.implements_protocol(cls, protocol)
    assert result is expected



import typing as t
from typing import Optional

from pybus.base.handlers import HandlerWrapper
from pybus.base.routers.eventrouter import EventRouter
from pybus.base.routers.requestrouter import RequestRouter
from pybus.core import exceptions as exc
from pybus.core.api.broker import AbstractBrokerAdapter
from pybus.core.api.dispatcher import DispatcherProtocol
from pybus.core.api.typing import HandlerType, MessageType
from pybus.core.types import EMPTY


class Dispatcher(DispatcherProtocol[EventRouter, RequestRouter, HandlerWrapper]):
    """Base dispatcher protocol implementation."""

    def __init__(
        self,
        events_router_cls: type[EventRouter] = EventRouter,
        commands_router_cls: Optional[type[RequestRouter]] = RequestRouter,
        queries_router_cls: Optional[type[RequestRouter]] = None,
        broker: Optional[AbstractBrokerAdapter] = None,
    ) -> None:
        self._events = events_router_cls(self)
        self._commands = commands_router_cls(self) if commands_router_cls is not None else None
        self._queries = queries_router_cls(self) if queries_router_cls is not None else None

        self._broker = broker
        self._started = False

    @property
    def events(self):  # pragma: no cover
        ...

    @events.getter
    def events(self) -> EventRouter:
        """Return events proxy"""
        if self._events is None:
            raise exc.ImproperlyConfigured(
                reason="Dispatcher Commands router has not been initialized"
            )
        return self._events

    @property
    def commands(self):  # pragma: no cover
        ...

    @commands.getter
    def commands(self) -> RequestRouter:
        """Return commands proxy"""
        if self._commands is None:
            raise exc.ImproperlyConfigured(
                reason="Dispatcher Commands router has not been initialized"
            )
        return self._commands

    @property
    def queries(self):  # pragma: no cover
        ...

    @queries.getter
    def queries(self) -> RequestRouter:
        """Return queries proxy"""
        if self._queries is None:
            raise exc.ImproperlyConfigured(
                reason="Dispatcher Queries router has not been initialized"
            )
        return self._queries

    @property
    def is_started(self) -> bool:
        """Return True if dispatcher has started."""
        return self._started

    def register_event_handler(  # noqa
        self,
        message: MessageType,
        handler: HandlerType,
        argname: t.Optional[str] = EMPTY,
        **initkwargs,
    ) -> HandlerWrapper:
        return self.events.bind(message, handler, argname, **initkwargs)

    def register_command_handler(  # noqa
        self,
        message: MessageType,
        handler: HandlerType,
        argname: t.Optional[str] = EMPTY,
        **initkwargs,
    ) -> HandlerWrapper:
        return self.commands.bind(message, handler, argname, **initkwargs)

    def register_query_handler(  # noqa
        self,
        message: MessageType,
        handler: HandlerType,
        argname: t.Optional[str] = EMPTY,
        **initkwargs,
    ) -> HandlerWrapper:
        return self.queries.bind(message, handler, argname, **initkwargs)

    def start(self) -> None:
        """Start the dispatcher."""
        if self._events is None:
            raise exc.ImproperlyConfigured(
                reason=f'{self.__class__.__name__} needs at least event router to work'
            )

        self.events.setup(broker=self._broker, name='events')

        if self._commands:
            self.commands.setup(broker=self._broker, name='commands')

        if self._queries:
            self.queries.setup(broker=self._broker, name='queries')

        self._started = True


default_dispatcher = Dispatcher()



from base.dispatcher import Dispatcher as Dispatcher
from base.dispatcher import default_dispatcher as dispatcher
from base.engine.events import EventEngine as EventEngine
from base.engine.requests import RequestEngine as RequestEngine
from base.handlers import AbstractHandler as AbstractHandler
from base.maps import EventHandlerMap as EventHandlerMap
from base.maps import RequestHandlerMap as RequestHandlerMap
from base.routers.eventrouter import EventRouter as EventRouter
from base.routers.requestrouter import RequestRouter as RequestRouter

__all__ = [
    "AbstractHandler",
    "Dispatcher",
    "EventHandlerMap",
    "EventEngine",
    "EventRouter",
    "RequestHandlerMap",
    "RequestEngine",
    "RequestRouter",
    "dispatcher",
]



import abc
import inspect
from logging import getLogger

from pybus.core import helpers
from pybus.core.api.handlers import AbstractHandler, AbstractHandlerWrapper
from pybus.core.api.typing import MessageType
from pybus.core.types import EMPTY

logger = getLogger(__name__)


class PyBusAbstractHandler(AbstractHandler, metaclass=abc.ABCMeta):
    """
    PyBusHandler abstract base class.

    To implement:
      - handle

    """

    def __init__(self):
        self._events = []

    async def add_event(self, event: MessageType):
        """Add event to emit later."""
        self._events.append(event)

    async def dump_events(self):
        """Return list of collected events."""
        events = self._events
        self._events = []
        return events


class HandlerWrapper(AbstractHandlerWrapper):
    """Handler wrapper."""

    def __repr__(self):
        return helpers.handler_repr(self._handler)

    __str__ = __repr__

    async def handle(self, message: MessageType):
        """Handle message."""
        handler = self._handler.handle if hasattr(self._handler, 'handle') else self._handler

        if not self._inject:
            result = await handler()
        elif self._argname is not EMPTY:
            result = await handler(**{self._argname: message})
        else:
            result = await handler(message)

        if result and not hasattr(self._handler, 'add_event'):
            await self.add_event(result)

        return result

    async def can_handle(self, message: MessageType) -> bool:
        """Return True if handler can handle the given message."""
        if hasattr(self._handler, 'can_handle'):
            return await self._handler.can_handle(message)

        ret = self._message is not EMPTY and (
            message == self._message
            or (
                inspect.isclass(self._message)
                and (isinstance(message, self._message) or issubclass(message, self._message))
            )
        )

        return ret

    async def add_event(self, event: MessageType):
        """Add event to emit later."""
        if hasattr(self._handler, 'add_event'):
            return await self._handler.add_event(event)

        self._events.append(event)

    async def dump_events(self):
        """Return list of collected events."""
        if hasattr(self._handler, 'dump_events'):
            return await self._handler.dump_events()

        events = self._events
        self._events = []
        return events



import typing as t


@t.runtime_checkable
class DescriptorProtocol(t.Protocol):
    """Descriptor protocol."""

    def __get__(self, instance, owner): ...

    def __set__(self, instance, value): ...

    def __delete__(self, instance): ...



from pybus.base.maps import EventHandlerMap
from pybus.core.api.engine import AbstractEngine
from pybus.core.api.typing import MessageType, PyBusWrappedHandler


class EventEngine(AbstractEngine[EventHandlerMap]):

    async def handle(self, handlers: frozenset[PyBusWrappedHandler], message: MessageType):
        new_events = []
        for handler in handlers:
            await handler.handle(message)
            new_events.extend(await handler.dump_events())

        return new_events

    async def handle_side_events(self, *events: MessageType) -> None:
        """Handle side handler event"""
        for event in events:
            await self.put_to_queue(event)



import abc
import asyncio
from logging import getLogger
from typing import Generic, Optional

from pybus.core.api.broker import AbstractBrokerAdapter
from pybus.core.api.typing import MapReturnType, MessageMapType, MessageType

logger = getLogger(__name__)


class AbstractEngine(Generic[MessageMapType], metaclass=abc.ABCMeta):
    """Engine protocol.

    Implement:
      - handle
    """

    def __init__(self, message_map: MessageMapType):
        self._map = message_map
        self._queue = asyncio.Queue()
        self._broker: Optional[AbstractBrokerAdapter] = None
        self._workers: int = 0
        self._name: Optional[str] = None

        self._started = False

    @property
    def is_started(self) -> bool:
        """Return True if the engine is started."""
        return self._started

    def set_name(self, name: str):
        """Set the engine name."""
        self._name = name or self.__class__.__name__

    async def put_to_queue(self, message: MessageType) -> None:
        """Send message"""
        await self._queue.put(message)

    async def _worker(self):
        self._workers += 1
        worker = f"{self._name!r} worker #{self._workers}"
        logger.debug(f"{worker} started.")

        while True:
            try:
                logger.debug(f'{worker} waiting for messages...')
                message = await self._queue.get()
                logger.debug(f'{worker} got message: {message!r}')

                handler = self._map.find(message)

                if new_events := await self.handle(handler, message):
                    logger.debug(f'{worker} got new events: {new_events} from {handler!r}')
                    await self.handle_side_events(*new_events)

                else:
                    logger.debug(f'{worker} no new events from {handler!r}')

            except asyncio.CancelledError:
                break

            except Exception as e:
                await self.handler_error(e)

    async def _broker_loop(self) -> None:
        while True:
            try:
                message = await self._broker.get()
                await self.put_to_queue(message)
            except asyncio.CancelledError:
                break

    async def handler_error(self, error: str | Exception) -> None:  # noqa
        """Error handler"""
        logger.exception(error)

    @abc.abstractmethod
    async def handle(self, handler: MapReturnType, message: MessageType) -> list[MessageType]:
        """Handler the message. Optionally can return new messages."""

    @abc.abstractmethod
    async def handle_side_events(self, *events: MessageType) -> None:
        """Handle handler event"""

    def setup_broker(self, broker: AbstractBrokerAdapter):
        """Setup broker loop."""
        if broker is None:
            return logger.error('broker was not provided')

        self._broker = broker
        tasks = [self._broker_loop() for _ in range(broker.get_workers())]
        asyncio.gather(*tasks)

        # additional workers for broker queue
        self.start(workers=broker.get_workers())

    def start(self, workers: int = 3) -> None:
        """Start the engine."""
        tasks = [self._worker() for _ in range(workers)]
        asyncio.gather(*tasks)

        self._started = True



from pybus.base.engine.requests import RequestEngine
from pybus.base.maps import RequestHandlerMap
from pybus.base.routers.messagerouter import AbstractBaseMessageRouter


class RequestRouter(AbstractBaseMessageRouter[RequestEngine]):
    map_cls = RequestHandlerMap
    engine_cls = RequestEngine

    def get_engine(self):
        return self.engine_cls(event_engine=self._dispatcher.events.engine, message_map=self._map)



from pybus.base.engine.events import EventEngine
from pybus.base.maps import EventHandlerMap
from pybus.base.routers.messagerouter import AbstractBaseMessageRouter


class EventRouter(AbstractBaseMessageRouter[EventEngine]):
    map_cls = EventHandlerMap
    engine_cls = EventEngine

    def get_engine(self):
        return self.engine_cls(message_map=self._map)



import inspect
from typing import Any

from pybus.core.types import ProtocolType

isfun = inspect.isfunction
iscoro = inspect.iscoroutinefunction
getsig = inspect.signature


def has_self(func):
    """Return True if the given function has self attribute."""
    spec = inspect.getfullargspec(func)
    maybehas = len(spec) > 1 and spec[0] in ["cls", 'self']
    return maybehas or inspect.ismethod(func) and not isinstance(func, staticmethod)


def implements_protocol(cls: Any, protocol: ProtocolType):
    """Return True if the given object implements the protocol."""
    if not isinstance(cls, protocol):
        return False

    for name, member in inspect.getmembers(protocol):
        if name.startswith("__"):
            continue

        protocol_member = getattr(protocol, name)
        if not (isfun(protocol_member) or isfun(protocol_member)):
            continue

        class_member = getattr(cls, name, None)
        if iscoro(protocol_member) != iscoro(class_member):
            return False

        if getsig(protocol_member) != getsig(class_member):
            return False

    return True



import abc
import typing as t

from pybus.base.handlers import HandlerWrapper
from pybus.core.api.routers import AbstractMessageRouter
from pybus.core.api.typing import EngineType, HandlerType, MessageType
from pybus.core.inspection import sig


class AbstractBaseMessageRouter(
    AbstractMessageRouter[EngineType, HandlerWrapper],
    t.Generic[EngineType],
    metaclass=abc.ABCMeta
):
    """Base class for event and request router implementations."""

    def get_meta(
        self,
        message: MessageType,
        handler: HandlerType,
        argname: t.Optional[str],
        **initkwargs,
    ) -> sig.HandlerMetaData:
        """Return Handler meta"""
        return sig.check_signature(handler, message, argname, **initkwargs)



def handler_repr(handler):  # pragma: no cover
    """Return a string representation of the given handler."""
    cls = (
        handler.__self__.__class__.__name__ if hasattr(handler, "__self__")
        else handler.__class__.__name__ if hasattr(handler, "__class__")
        else ""
    )

    div = "." if cls else ""
    handler_name = f'{handler.__module__}.{cls}{div}'

    if hasattr(handler, '__name__'):
        handler_name += handler.__name__

    else:
        if "handle" in dir(handler):
            handler_name += 'handle'
        else:
            print(f'unknown handler {handler=!r}')
            print(dir(handler))

    return handler_name



from pybus.core import helpers

from . import __library_name__
from .api.typing import HandlerType, MessageType


class PyMessagesException(Exception):
    """base Exception"""
    reason: str = ""

    def __init__(self, message=None, **kwargs):
        for key, value in kwargs.items():
            if key == "handler" and not isinstance(value, str):
                value = helpers.handler_repr(value)

            setattr(self, key, value)

        self.message = message

    def get_message(self):  # pragma: no cover
        """Return the message for this exception"""
        return self.__doc__

    def with_reason(self, reason):
        """Return error with provided reason"""
        return self.__class__(**self.__dict__, reason=reason)

    def __str__(self):
        reason = f": {self.reason}" if self.reason else ""
        return f'{self.get_message()}{reason}'


class ImproperlyConfigured(PyMessagesException):
    """library configuration error."""

    def get_message(self):  # pragma: no cover
        return f'{__library_name__} configuration error'


class HandlerDoesNotExist(ImproperlyConfigured):
    """handler does not exist."""
    pymessage: MessageType

    def get_message(self):  # pragma: no cover
        return f'handler for {self.pymessage!r} does not exist'


class HandlerIsSync(ImproperlyConfigured):
    """handler is sync."""
    handler: HandlerType

    def get_message(self):  # pragma: no cover
        return f"handler {self.handler!r} is sync"


class HandlerSignatureError(ImproperlyConfigured):
    """handler signature error."""
    handler: HandlerType

    def get_message(self):  # pragma: no cover
        return f'handler {self.handler!r} has invalid signature'



import typing as t

from pybus.core.api.typing import (
    EventRouterType,
    HandlerType,
    MessageType,
    PyBusWrappedHandler,
    RequestRouterType,
)
from pybus.core.types import EMPTY


@t.runtime_checkable
class DispatcherProtocol(t.Protocol[EventRouterType, RequestRouterType, PyBusWrappedHandler]):
    """DispatcherProtocol protocol."""

    @property
    def events(self) -> EventRouterType:
        """Return events proxy"""

    @property
    def commands(self) -> RequestRouterType:
        """Return commands proxy"""

    @property
    def queries(self) -> RequestRouterType:
        """Return queries proxy"""

    @property
    def is_started(self) -> bool:
        """Return True if dispatcher has started."""

    def register_event_handler(
        self,
        message: MessageType,
        handler: HandlerType,
        argname: t.Optional[str] = EMPTY,
        **initkwargs,
    ) -> PyBusWrappedHandler:
        """Register event handler."""

    def register_command_handler(
        self,
        message: MessageType,
        handler: HandlerType,
        argname: t.Optional[str] = EMPTY,
        **initkwargs,
    ) -> PyBusWrappedHandler:
        """Register command handler."""

    def register_query_handler(
        self,
        message: MessageType,
        handler: HandlerType,
        argname: t.Optional[str] = EMPTY,
        **initkwargs,
    ) -> PyBusWrappedHandler:
        """Register query handler."""



import abc
import functools
import typing as t
from collections.abc import Callable
from typing import Optional

from pybus.core.api.broker import AbstractBrokerAdapter
from pybus.core.api.dispatcher import DispatcherProtocol
from pybus.core.api.handlers import HandlerMetaDataProtocol
from pybus.core.api.maps import AbstractHandlerMap
from pybus.core.api.typing import (
    EngineType,
    HandlerType,
    MessageType,
    PyBusWrappedHandler,
)
from pybus.core.types import EMPTY


class AbstractMessageRouter(
    t.Generic[EngineType, PyBusWrappedHandler],
    metaclass=abc.ABCMeta,
):
    """Router protocol."""
    map_cls: type[AbstractHandlerMap]
    engine_cls: type[EngineType]

    def __init__(self, dispatcher: DispatcherProtocol):
        self._dispatcher = dispatcher
        self._map = self.get_map()
        self._engine: Optional[EngineType] = None

    async def send(self, message: MessageType) -> None:
        """Send message to the bus."""
        assert self._dispatcher.is_started, "you should start the dispatcher first"
        assert self._engine.is_started, "you should start the engine first"
        await self._engine.put_to_queue(message)

    @property
    def engine(self) -> EngineType:
        """Return the engine for this router."""
        return self._engine

    def get_map(self):
        """Returns the message handler map."""
        if self.map_cls is None:
            raise NotImplementedError()

        return self.map_cls()

    def bind(
        self,
        message: MessageType,
        handler: HandlerType,
        argname: t.Optional[str] = EMPTY,
        **initkwargs,
    ) -> PyBusWrappedHandler:
        """Bind handler for the given message."""
        meta = self.get_meta(message, handler, argname, **initkwargs)
        return self._map.wrap_handler(meta)

    def register(
        self,
        message: MessageType,
        argname: t.Optional[str] = EMPTY,
        **initkwargs,
    ) -> Callable:
        """Register handler for the given message as decorator."""

        def _wrapper(handler: HandlerType):
            wrapped_handler = self.bind(message, handler, argname, **initkwargs)

            @functools.wraps(handler)
            async def _decorated(*args, **kwargs):
                return await wrapped_handler.handle(*args, **kwargs)

            return _decorated

        return _wrapper

    @abc.abstractmethod
    def get_meta(
        self,
        message: MessageType,
        handler: HandlerType,
        argname: t.Optional[str],
        **initkwargs,
    ) -> HandlerMetaDataProtocol:
        """Return Handler meta"""

    @abc.abstractmethod
    def get_engine(self):
        """Set up the engine for this router."""

    def setup(self, name: Optional[str] = None, broker: Optional[AbstractBrokerAdapter] = None):
        """setup map, routes, ..."""
        self._map.build()

        self._engine = self.get_engine()
        self._engine.set_name(name)
        self._engine.setup_broker(broker=broker)

        self._engine.start()



import typing as t


def hashable(value: t.Any) -> t.TypeGuard[t.Hashable]:
    """Direct way to check hashable"""
    try:
        hash(value)
        return True
    except TypeError:
        return False


@t.runtime_checkable
class _O(t.Protocol):
    pass


ProtocolType = type(_O)

del _O

EMPTY = object()



import typing as t

if t.TYPE_CHECKING:
    from pybus.core.api.engine import AbstractEngine  # noqa
    from pybus.core.api.handlers import (  # noqa
        AbstractHandler,
        AbstractHandlerWrapper,
        HandlerMetaDataProtocol,
    )
    from pybus.core.api.maps import AbstractHandlerMap  # noqa
    from pybus.core.api.routers import AbstractMessageRouter  # noqa

Message: t.TypeAlias = type[t.Any] | t.Hashable
MessageType = t.TypeVar("MessageType", bound=Message)

P = t.ParamSpec("P")
PyBusHandler = t.TypeVar("PyBusHandler", bound="AbstractHandler")
PyBusWrappedHandler = t.TypeVar("PyBusWrappedHandler", bound="AbstractHandlerWrapper")

PyBusHandlerMeta = t.TypeVar("PyBusHandlerMeta", bound="HandlerMetaDataProtocol")

ReturnType = t.TypeVar("ReturnType", t.Any, None)
HandlerReturnType: t.TypeAlias = t.Awaitable[ReturnType]

HandlerType: t.TypeAlias = t.Union[
    PyBusHandler,
    PyBusWrappedHandler,
    t.Callable[[], HandlerReturnType],
    t.Callable[[MessageType], HandlerReturnType],
    t.Callable[[MessageType, P.kwargs], HandlerReturnType],
]

MessageMapType = t.TypeVar("MessageMapType", bound="AbstractHandlerMap")
MapReturnType = t.TypeVar("MapReturnType", HandlerType, frozenset[HandlerType])

EngineType = t.TypeVar("EngineType", bound="AbstractEngine")

EventRouterType = t.TypeVar('EventRouterType', bound="AbstractMessageRouter")
RequestRouterType = t.TypeVar('RequestRouterType', bound="AbstractMessageRouter")



from typing import NamedTuple, Optional, TypeVar

from pybus.base.handlers import HandlerWrapper
from pybus.core.api.maps import AbstractHandlerMap
from pybus.core.api.typing import (
    HandlerType,
    MapReturnType,
    MessageType,
    PyBusHandlerMeta,
)
from pybus.core.exceptions import HandlerDoesNotExist
from pybus.core.types import EMPTY

EmptySet = frozenset()
HandlerFrozenSet = TypeVar("HandlerFrozenSet", bound=frozenset[HandlerType])


class HandlerMetaData(NamedTuple):
    """Meta data for dispatcher"""
    inject: bool
    argname: Optional[str] = EMPTY
    message: Optional[MessageType] = EMPTY


class EventHandlerMap(AbstractHandlerMap[HandlerFrozenSet]):
    """Event handler map."""

    def add(self, message: MessageType, handler: HandlerType) -> None:
        if self._frozen:
            raise RuntimeError("map is frozen")

        self._storage.setdefault(self.make_key(message), set()).add(handler)

    def find(self, key: MessageType) -> MapReturnType:
        return self._storage.get(self.make_key(key), EmptySet)

    @classmethod
    def freeze(cls, val: set[HandlerType]) -> MapReturnType:
        return frozenset(val)

    def wrap_handler(self, meta: PyBusHandlerMeta) -> HandlerWrapper:
        wrapped = HandlerWrapper(meta=meta)
        self.add(message=meta.message, handler=wrapped)
        return wrapped


class RequestHandlerMap(AbstractHandlerMap[HandlerType]):
    """Request handler map."""

    def add(self, message: MessageType, handler: HandlerType) -> None:
        if self._frozen:
            raise RuntimeError("map is frozen")

        self._storage[self.make_key(message)] = handler

    def find(self, key: MessageType) -> MapReturnType:
        try:
            return self._storage[self.make_key(key)]
        except KeyError:
            raise HandlerDoesNotExist(pymessage=key)

    @classmethod
    def freeze(cls, val: HandlerType) -> MapReturnType:
        return val

    def wrap_handler(self, meta: PyBusHandlerMeta) -> HandlerWrapper:
        wrapped = HandlerWrapper(meta=meta)
        self.add(message=meta.message, handler=wrapped)
        return wrapped



from pybus.base.engine.events import EventEngine
from pybus.base.maps import RequestHandlerMap
from pybus.core.api.engine import AbstractEngine
from pybus.core.api.typing import MessageType, PyBusWrappedHandler


class RequestEngine(AbstractEngine[RequestHandlerMap]):

    def __init__(self, event_engine: EventEngine, message_map: RequestHandlerMap):
        super().__init__(message_map=message_map)
        self._event_engine = event_engine

    async def handle(self, handler: PyBusWrappedHandler, message: MessageType):
        await handler.handle(message)
        return await handler.dump_events()

    async def handle_side_events(self, *events: MessageType) -> None:
        """Handle side handler event"""
        for event in events:
            await self._event_engine.put_to_queue(event)



import inspect
from typing import Any, NamedTuple, Optional

from pybus.core import exceptions as exc
from pybus.core.api.handlers import HandlerProtocol
from pybus.core.api.typing import HandlerType, Message
from pybus.core.inspection import utils
from pybus.core.types import EMPTY

KIND_MAP = {
    "pos_only": inspect.Parameter.POSITIONAL_ONLY,
    "pos_or_kw": inspect.Parameter.POSITIONAL_OR_KEYWORD,
    "pos": inspect.Parameter.VAR_POSITIONAL,
    "kw_only": inspect.Parameter.KEYWORD_ONLY,
    "kw": inspect.Parameter.VAR_KEYWORD,
}

KWSPEC = ("pos_or_kw", "kw_only", "kw")


class HandlerMetaData(NamedTuple):
    """Meta data for dispatcher"""
    handler: HandlerType
    inject: bool
    argname: Optional[str] = EMPTY
    message: Optional[Message] = EMPTY


def check_signature(
    handler: HandlerType,
    message: Message = EMPTY,
    argname: Optional[str] = EMPTY,
    **initkwargs: dict[str, Any],
) -> HandlerMetaData:
    """Check signature of the handler"""
    if not handler:
        raise exc.HandlerDoesNotExist(pymessage=message)

    cls = handler
    if inspect.isclass(handler):
        cls = handler(**initkwargs)
        if utils.implements_protocol(cls, HandlerProtocol):
            handler = cls.handle
        else:
            if not hasattr(cls, "__call__"):
                raise exc.HandlerSignatureError(
                    handler=handler,
                    reason=f'class {cls.__class__.__name__} '
                           f'should implement `__call__` method or `HandlerProtocol`',
                )
            handler = cls.__call__

    if not inspect.iscoroutinefunction(handler):
        raise exc.HandlerIsSync(handler=cls, reason='accepted only awaitable')

    argposition = 0 if not utils.has_self(handler) else 1

    sig = inspect.signature(handler)
    param_amount = len(sig.parameters)

    if param_amount + argposition == argposition:
        # 1: существуют ли вообще аргументы
        # - если нет, то это норм, просто заканчиваем проверку,
        # а сообщение в хэндлер не передаем
        return HandlerMetaData(handler=handler, inject=False, message=message)

    error = exc.HandlerSignatureError(pymessage=message, handler=handler)
    reason = None
    is_valid = False
    # todo annotated bind
    for index, param in enumerate(sig.parameters.values()):

        if index == param_amount - 1:
            # достигли конца цикла, необходимо прервать
            # просто дождаться - не вариант, т.к. используется конструкция for...else
            is_valid = True
            break

        if argname is EMPTY and index > argposition and param.kind in KIND_MAP:
            reason = (f"argument must be the first parameter if argname is not specified "
                      f"and kind of argument in {', '.join(KWSPEC)!r}")
            # 2: если аргументов больше, чем argpos, но мы не знаем имя аргумента,
            # при этом тип аргумента среди именных, ошибка-мы не сможем найти аргумент для вставки,
            # предложить указать argname
            break

        if argname != param.name:
            # 3. Имя указано, но не соответствует аргументу, пропускаем
            continue

    else:
        # for-loop did not find compatible argument in the handler signature
        reason = f'argname {argname!r} not found'

    if is_valid is False:
        raise error.with_reason(reason=reason)

    return HandlerMetaData(handler=cls, inject=True, argname=argname, message=message)



import abc

from pybus.core.api.typing import MessageType


class AbstractBrokerAdapter(metaclass=abc.ABCMeta):
    """Broker adapter."""

    @abc.abstractmethod
    def get_workers(self) -> int:
        """Get number of workers."""

    @abc.abstractmethod
    async def get(self) -> MessageType:
        """Get message from the broker queue."""



import abc
import copy
import inspect
from types import MappingProxyType
from typing import Generic

from pybus.core.api.handlers import AbstractHandlerWrapper
from pybus.core.api.typing import (
    HandlerType,
    MapReturnType,
    MessageType,
    PyBusHandlerMeta,
    PyBusWrappedHandler,
)


class AbstractHandlerMap(Generic[MapReturnType], metaclass=abc.ABCMeta):
    """Handler map.

    Abstract methods:
      - add
      - find
      - freeze
      - wrap_handler
    """

    def __init__(self):
        self._storage = {}
        self._frozen = False

    @property
    def frozen(self):
        """Return frozen state"""
        return self._frozen

    @classmethod
    def make_key(cls, message: MessageType) -> MessageType:
        """Return message key"""
        m_type = message if inspect.isclass(message) else type(message)
        return m_type if not issubclass(m_type, str) else message

    def build(self):
        """Build map. Handlers would be frozen after."""
        if self._frozen:
            raise RuntimeError("map is frozen")

        map_keys = list(self._storage.keys())
        map_copy = copy.copy(self._storage)
        self._storage.clear()

        for m_key in map_keys:
            self._storage[m_key] = self.freeze(map_copy[m_key])

        self._storage = MappingProxyType(self._storage)
        self._frozen = True

    def can_handle(self, message: MessageType) -> bool:
        """Return True if the given message key can be handled."""
        return self.make_key(message) in self._storage

    @abc.abstractmethod
    def add(self, message: MessageType, handler: PyBusWrappedHandler) -> None:
        """Add handler to the map."""

    @abc.abstractmethod
    def find(self, key: MessageType) -> MapReturnType:
        """Return handler or handlers for the given key."""

    @abc.abstractmethod
    def wrap_handler(self, meta: PyBusHandlerMeta) -> AbstractHandlerWrapper:
        """Wrap handler"""

    @classmethod
    @abc.abstractmethod
    def freeze(cls, val: HandlerType | set[HandlerType]) -> MapReturnType:
        """Freeze map value."""
